/*--------------------------------------------------------------------------*/
/* system dependent functions and operations used by the test program */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include "sortsys.h"

/*--------------------------------------------------------------------------*/
/* system dependent initializations */
/*--------------------------------------------------------------------------*/

#ifdef THINK_C
  #include <console.h>
#endif

void sysinit(void)
{
  /* nothing to do */
}

/*--------------------------------------------------------------------------*/
/* system information */
/*--------------------------------------------------------------------------*/

#if SYS_MACINTOSH

#include <GestaltEqu.h>
#include <ToolUtils.h>
#include <strings.h>

/* get information about a macintosh computer */
void sysinfomac(char *mc_name, char *os_name, char *cc_name)
{
  long machineType;
  long systemVersion;
  
  /* get compiler version */
  #ifdef THINK_C
    #if THINK_C > 1
      sprintf(cc_name, "THINK C v%d", THINK_C);
    #elif THINK_C == 1
      sprintf(cc_name, "THINK C v4");
    #else
      sprintf(cc_name, "THINK C", THINK_C);
    #endif
  #elif MPW
    sprintf(cc_name, "MPW");
  #elif __MWERKS__
    sprintf(cc_name, "MetroWerks");
  #endif
    
  /* get machine type */
  if (Gestalt(gestaltMachineType, &machineType) == noErr) {
    GetIndString((unsigned char *) mc_name, kMachineNameStrID, machineType);
    p2cstr((unsigned char *) mc_name);
  }
    
  /* get system version */
  if (Gestalt(gestaltSystemVersion, &systemVersion) == noErr) {
    sprintf(os_name, "System %lx.%lx.%lx",
    	(systemVersion >> 8) & 0x00FF,
    	(systemVersion >> 4) & 0x000F,
    	(systemVersion >> 0) & 0x000F);
  }
}

#endif /* SYS_MACINTOSH */

/*--------------------------------------------------------------------------*/
/* system dependent interface to profiler */
/*--------------------------------------------------------------------------*/

#ifndef PROFILE
  #define PROFILE (0)
#endif

#ifdef THINK_C
  #include <profile.h>
#endif

#if SYS_UNIX && PROFILE
  /* These are just prototypes for the profiler calls; I couldn't find
     the actual header file for these functions. */
  void moncontrol(int mode);
  void monitor(int);
#endif

/* initialize profiler */
void profile_init(void)
{
  #if PROFILE
    #if SYS_UNIX
    #elif THINK_C
      InitProfile(200, 200);
    #endif
  #endif
}

/* start profiling */
void profile_start(void)
{
  #if PROFILE
    #if SYS_UNIX
      moncontrol(TRUE);
    #endif
  #endif
}

/* stop profiling */
void profile_stop(void)
{
  #if PROFILE
    #if SYS_UNIX
      moncontrol(FALSE); /* disable profiler */
    #elif THINK_C
      _profile = FALSE;
    #endif
  #endif
}

/* output profile info */
void profile_dump(void)
{
  #if PROFILE
    #if SYS_UNIX
      /* create file "mon.out" */
      monitor(FALSE);
    #elif THINK_C
      DumpProfile();
    #endif
  #endif
}
